#ifndef BSETS2_ARCHIVE_H
#define BSETS2_ARCHIVE_H

#include "options.h"
#include "PointSet.hpp"
#include "Group.hpp"


////////////////////////////////////////
// DECLARATIONS
////////////////////////////////////////

/**
 * Member of an Archive object. Do not use directly
 */
class Member {

private:
    // point set stored in this archive, may be in canonical form or not depending on canonical
    PointSet pointSet;

    // whether this member is in canonical form or not
    bool canonical;

public:
    // invariant for this point set
    const int invariant;

    /**
     * New member with a copy of the given set.
     */
    Member (int invariant, PointSet &set) : invariant(invariant), canonical(false), pointSet(set) {}

    /**
     * Return the canonical form of this member. If necessary, compute and replace the stored
     * set by its canonical form
     */
    PointSet &canonicalForm (const Group &group);

};


/**
 * Represents a set of non-equivalent point sets of the same size. Used during generation to
 * check whether a newly generated set is equivalent to a set that was generated earlier.
 *
 * Uses invariants to avoid unnecessary computation of the canonical form of its members.
 *
 */
class Archive {

private:

    // size of this archive
    int size = 0;

    // pointers to members
    Member *member[NOP]{};   // not initialized!

    // reference to the group
    const Group &group;

public:
    
    explicit Archive (const Group &group) : group (group) {}

    // cleans up memory used
    ~Archive();

    /**
     * Add the given set (with given invariant) to the archive.
     * Returns whether the set was added
     */
    bool add (int invariant, PointSet &set);

private:
    // is there a member with the given invariant
    bool hasInvariant (int invariant);

};

////////////////////////////////////////
// DEFINITIONS
////////////////////////////////////////

PointSet & Member::canonicalForm(const Group & group) {
    if (! canonical) {
        PointSet setCopy = pointSet;
        group.canonicalMap(setCopy, pointSet);
        canonical = true;
    }
    return pointSet;
}

bool Archive::hasInvariant(int invariant) {
    for (int i=0; i < size; i++) {
        if (member[i] -> invariant == invariant) {
            return true;
        }
    }
    return false;
}

// release memory
Archive::~Archive () {
    for (int i=0; i < size; i++) {
        delete member[i];
    }
}

bool Archive::add(int invariant, PointSet &set) {
    // tentatively add to the archive - avoids an unnecessary copy of the canonical form later
    member[size] = new Member(invariant, set) ;
    if (hasInvariant(invariant)) {
        PointSet &newCanForm = member[size]->canonicalForm(group);
        for (int i=0; i < size; i++) {
            if (member[i]->canonicalForm(group) == newCanForm) {
                delete member[size]; // not needed after all...
                return false;
            }
        }
    }
    // really add to the archive
    size ++;
    return true;
}


#endif  //BSETS2_GROUP_H