// Note: this file is part of a header-only library. See Readme.md

#ifndef BSETS2_FF_H
#define BSETS2_FF_H

#include "options.h"

////////////////////////////////////////
// DECLARATIONS
////////////////////////////////////////

class FFel {

private:
    char value; // smaller than int

private:
    // constructor for internal use only, parameter v must be in the range 0..Q-1
    FFel(char v, bool not_used) : value(v) {}

public:

    /**
     * Converts an integer to a finite field element representing the same value. Can be used
     * as implicit conversion
     */
    FFel(char v); // NOLINT(google-explicit-constructor)

    FFel() = default;

    FFel operator+(const FFel &other) const {
        return addition[value][other.value];
    }

    FFel operator-(const FFel &other) const {
       return subtraction[value][other.value];
    }

    FFel operator*(const FFel &other) const {
        return multiplication[value][other.value];
    }

    FFel inverse() const {
        return division[1][value];
    }

    FFel operator/(const FFel &other) const {
       return division[value][other.value];
    }

    bool operator==(const FFel &other) const {
        return value == other.value;
    }

    bool operator!=(const FFel &other) const {
        return value != other.value;
    }

    // should be safe to use  == 0 instead
    bool isZero() const {
        return value == 0;
    }

    // should be safe to use == 1 instead
    bool isOne() const {
        return value == 1;
    }

    /**
     * Convert to integer. Important! This only works for prime fields
     */
    explicit operator int() const {
        return value; // note that this only works for prime fields!
    }

    /**
     * Return the internal index used to represent this element. We have elements[el.index()] == el
     */
    int index() const {
        return value;
    }

    /**
     * All elements of the field
     */
    static const FFel elements[Q];

    static const FFel zero;

    static const FFel one;

private:

    // lookup tables for elementary operations, unfortunately not const as they need to be initialized later
    static FFel multiplication[Q][Q];
    static FFel division[Q][Q];
    static FFel addition[Q][Q];
    static FFel subtraction[Q][Q];

    static bool initialize();
    static bool initialized; // hack for ensuring that the addition/subtraction/multiplication/division tables are initialized
};

////////////////////////////////////////
// ORDER SPECIFIC
////////////////////////////////////////

#if ORDER == 3
#include "ff/ff3.h"
#elif ORDER == 4
#include "ff/ff4.h"
#elif ORDER == 5
#include "ff/ff5.h"
#elif ORDER == 7
#include "ff/ff7.h"
#elif ORDER == 8
#include "ff/ff8.h"
#endif

////////////////////////////////////////
// DEFINITIONS
////////////////////////////////////////

const FFel FFel::zero = FFel(0, true);
const FFel FFel::one = FFel(1, true);

// Initialized by 'initialize()'
FFel FFel::multiplication[Q][Q] {};
FFel FFel::division[Q][Q] {} ;
FFel FFel::addition[Q][Q] {} ;
FFel FFel::subtraction[Q][Q] {} ;

bool FFel::initialize() {
    // addition + subtraction
    for (int i = 0; i < Q; i++) {
        for (int j=0; j < Q; j++) {
            addition[i][j] = FFel(add_table[i][j], true);
            subtraction[add_table[i][j]][j] = FFel(i, true);
        }
    }
    // multiplication + division
    for (int i = 1; i < Q; i++) {
        for (int j=1; j < Q; j++) {
            multiplication[i][j] = FFel(mul_table[i][j], true);
            division[mul_table[i][j]][j] = FFel(i, true);
        }
    }
    return true; // This line was missing when I merged Kris's cleanup.
}

bool FFel::initialized = initialize();

FFel::FFel(char v) {
    if (v==0) {
        value = 0;
    } else if (v==1) {
        value = 1;
    } else if (v==-1) {
        value = CHARACTERISTIC - 1;
    } else if (v > 0) {
        value = v % CHARACTERISTIC;
    } else { // if v < 0
        value = CHARACTERISTIC - v % CHARACTERISTIC;
    }
}

#endif //BSETS2_FF_H
