// Note: this file is part of a header-only library. See Readme.md

#ifndef BSETS2_GENERATOR_H
#define BSETS2_GENERATOR_H

#include <array>
#include <vector>

#include "options.h"

#include "PointSet.hpp"
#include "Group.hpp"
#include "OrbitPartition.hpp"
#include "Archive.hpp"

////////////////////////////////////////
// DECLARATIONS
////////////////////////////////////////

/*
 * Generates all blocking sets.
 */
class Generator {

public:
    int nrMinBlockingSetsOfSize[NOP + 1] = {0,};

    Generator();

    void generate();

    void processArgs(int argc, char *argv[]);

private:
    // Parameters to use when splitting process for parallellisation
    int splitDepth = NOP + 1; // default = do not split
    int splitIndex = 0;
    int splitModulus = 0;
    int splitCounter = 0;

private:
    const Group group; // PGL(3,q)

    PointSet set;  // current set being generated

    bool isSpecial(int pt, OrbitPartition &orbitsSet) const;

//    PointSet debugSet;
};

////////////////////////////////////////
// DEFINITIONS
////////////////////////////////////////

void Generator::processArgs(int argc, char *argv[]) {
    if (argc > 1) {
        if (argc != 4) {
            std::cerr << "Usage: " << argv[0] << " depth index modulus";
        } else {
            splitDepth = (int) strtol(argv[1], nullptr, 10);
            splitIndex = (int) strtol(argv[2], nullptr, 10);
            splitModulus = (int) strtol(argv[3], nullptr, 10);
        }
    }
}

void Generator::generate() {

    // bail out when not in the correct search branch (when parallelized)
    if (set.size == splitDepth) {
        if (splitCounter++ % splitModulus != splitIndex) {
            return;
        }
    }

    // report minimal blocking set
    if (set.size >= LOWERBOUNDB) {
        if (set.isBlockingSet()) {
            nrMinBlockingSetsOfSize[set.size]++; // We generate in a way that weeds out sets with redundant points. No redundant points + blocking all lines = minimal blocking set
            set.ship(std::cout);
            return; // If you found a minimal blocking set; there's no use in trying to expand it to ... a minimal blocking set?
        }
    }

    // recursion
    if (set.size < UPPERBOUNDMINB) {
        // we try to extend the current set with one extra point.
        Archive archive{group};

        // loop over all external points
        for (int pt = 0; pt < NOP; ++pt) {
            if (!set.present[pt]) {
                // check combinatorial properties (no full line, no redundant points)
                if (set.addAndCheck(pt)) {
                    // check whether added point is canonical
                    set.updateInvariants();

                    int result = set.checkLargestUniqueInvariant(pt);
                    if (result > 0) {
                        // this is the canonical point, as it has the biggest unique invariant
                        // add the set to the archive and continue
                        if (archive.add(set.computeInvariant(), set)) {
                            generate();
                        }
                    } else if (result == 0) {
                        // there are no unique invariants so the canonical pseudo-orbit is the one with highest invariant

                        int count = set.checkLargestInvariant(pt);
                        if (count == 1) {
                            // this should not happen
                            std::cerr << "Error in set.checkLargestUniqueInvariant, use debugger to find what went wrong" << std::endl;
                            exit(1);
                        } else if (count > 1) { // pseudo-orbit may split into multiple orbits
                            auto stabGroup = group.stabilizer(set);
                            OrbitPartition newOrbits{stabGroup};
                            if (isSpecial(pt, newOrbits)) { // alright, point is special, add set to archive.
                                if (archive.add(set.computeInvariant(), set)) {
                                    generate();
                                }
                            }
                        } // else: not the biggest invariant
                    } // else: not the biggest unique invariant
                }
                set.remove(pt);
            }
        }
    }
}

bool Generator::isSpecial(int pt, OrbitPartition &orbitsSet) const {
    // special case where quasi-orbit may split into multiple orbits
    //new implementation: orbits of CF are built from old orbits and canonicalMap
    PointSet newSet = set; // making a copy, necessary for method canonicalMap.
    int indexCM = group.canonicalMap(set, newSet);
    const Matrix &canonicalMap = group.list[indexCM];
    const Matrix &canonicalMapInv = canonicalMap.adjoint();
    int newPt = canonicalMap(Point::list[pt]).index();
    int newOrbitNr = orbitsSet.orbitArray[pt]; //new orbit nummer = oud orbit nummer
    int newInvariant = newSet.invariants[newPt]; // =set.invariants[pt] ?
    for (int i = 0; i < newPt; ++i) {
        if (newSet.present[i] && newSet.invariants[i] == newInvariant) {
            bool b = orbitsSet.orbitArray[canonicalMapInv(Point::list[i]).index()] == newOrbitNr;
            return b; // orbit with same invariant was encountered
        }
    }
    return true; // newPt was encountered first
}


Generator::Generator() {
    set.addAndCheck(0); //wlog
    set.addAndCheck(1); //wlog
    // Note that adding a further point collinear with the first two only works with certain invariants
    // It would be an 'optimization' that made hardly any difference anyway.

    set.updateInvariants();

}

#endif //BSETS2_GENERATOR_H
