// Note: this file is part of a header-only library. See Readme.md

#ifndef BSETS2_GROUP_H
#define BSETS2_GROUP_H


#include <vector>

#include "options.h"

#include "Matrix.hpp"
#include "PointSet.hpp"

////////////////////////////////////////
// DECLARATIONS
////////////////////////////////////////

class Group {
private:
    int IdMap;

public:
    Group();

    // Returns the stabilizer group of @param point set.
    std::vector<const Matrix *> stabilizer(const PointSet &set) const;

    /**
     * Compute the canonical form of the given set.
     * @param set [in] Point set for which the canonical form must be calculated
     * @param canonicalSet [out] Resulting canonical form. <b>Important!</b> Must have (a copy of) set as an initial value.
     * @return the index of the group element which mpas set onto canonical set
     */
    int canonicalMap(const PointSet &set, PointSet &canonicalSet) const;

public:
    Matrix list[GROUP_SIZE]; // all elements of this group

    //sgsList[i] is an array of all matrices that stabilize the point with index i.
    Matrix *sgsList[NOP][SGS_SIZE] = {nullptr,}; // sgs = Stabilizer Group Singleton

    //hlmList[i] is an array of all matrices that map the line with index i to the line with index Q+1 (special line that contains the points 0,1,...,Q)
    //Actually, this array contains the inverses of all those matrices (I will only be using this list to determine canonicalMap, where inverse matrices are desired)
    //Actually, it contains the index of the inverses of the matrices...
    int hlmList[NOP][SGS_SIZE] = {0,}; // hlm = Heavy Line Mapping.
};

////////////////////////////////////////
// DEFINITIONS
////////////////////////////////////////

std::vector<const Matrix *> Group::stabilizer(const PointSet &set) const {
    std::vector<const Matrix *> result;
    int uniquePt = set.hasUniqueInvariant();
    if (uniquePt > -1) {
        for (const Matrix *matrix : sgsList[uniquePt]) {
            if (set.isFixedBy(*matrix)) {
                result.push_back(matrix);
            }
        }
    } else {
        for (const Matrix &matrix : list) {
            if (set.isFixedBy(matrix)) {
                result.push_back(&matrix);
            }
        }
    }
    return result;
}


Group::Group() {

    if (Q == 3) { IdMap = 468; }
    if (Q == 5) { IdMap = 12400; } //TODO: wordt dit wel gebruikt? Might be a relic of an ancient time...
    if (Q == 7) { IdMap = 100548; }
    if (Q == 8) { IdMap = 228928; } // In general: IdMap has index q^2*(q-1)^2*NOP

    // and proceed with the actual initialization
    const FFel *FF = FFel::elements;

    int counter = 0;
    /*                         001
        Matrices of the form   abe
                               cdf
     */
    for (const auto &a : FFel::elements) {
        for (const auto &b : FFel::elements) {
            for (const auto &c : FFel::elements) {
                for (const auto &d : FFel::elements) {
                    if (a * d != b * c) {
                        for (const auto &e : FFel::elements) {
                            for (const auto &f : FFel::elements) {
                                list[counter] = Matrix(FFel::zero, FFel::zero, FFel::one, a, b, e, c, d, f);
                                counter++;
                            }
                        }
                    }
                }
            }
        }
    }

    /*                      01i           1ij
     Matrices of the form   jkl    and    klm
                            mno           nop
     */
    for (const auto &i : FFel::elements) {
        for (const auto &j : FFel::elements) {
            for (const auto &k : FFel::elements) {
                for (const auto &l : FFel::elements) {
                    for (const auto &m : FFel::elements) {
                        for (const auto &n : FFel::elements) {
                            for (const auto &o : FFel::elements) {
                                Matrix mat01i = Matrix(FFel::zero, FFel::one, i, j, k, l, m, n, o);
                                if (!mat01i.det().isZero()) {
                                    list[counter] = mat01i;
                                    counter++;
                                }
                                for (const auto &p : FFel::elements) {
                                    Matrix mat1ij = Matrix(FFel::one, i, j, k, l, m, n, o, p);
                                    if (!mat1ij.det().isZero()) {
                                        list[counter] = mat1ij;
                                        counter++;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    // TODO: where is the identity matrix?

/*
    for (int j = 0; j < GROUP_SIZE; ++j) {
        invList[j] = list[j].adjoint();
        // TODO can we please avoid to make another copy of the group?
        // Unfortunately, with the current algorithm, we do need a lookup table
        // otherwise the program slows down with at least a factor of 2
    }
*/

    PointSet set;
    for (int k = 0; k < NOP; ++k) {
        int myCounter = 0;
        set.add(k);
        set.updateInvariants();
        for (Matrix &matrix : list) { //TODO const?
            if (set.isFixedBy(matrix)) {
                sgsList[k][myCounter] = &matrix;
                ++myCounter;
            }
        }
        set.remove(k);
        set.updateInvariants();
    }

    int counters[NOP] = {0,};
    for (int n = 0; n < GROUP_SIZE; ++n) {
        Matrix &matrix = list[n];
        Point line = Point{matrix.m[0], matrix.m[3],
                           matrix.m[6]}; // the line that gets mapped to line Q+1 has this coordinate since line Q+1 has coordinate (1,0,0)
        int indexLine = line.index(); // current matrix maps this line to line Q+1
        hlmList[indexLine][counters[indexLine]] = n;
        counters[indexLine]++;
    }
}

// Computes the canonical form of the
int Group::canonicalMap(const PointSet &set, PointSet &canonicalSet) const {
    int bestMap = IdMap; // Initialize on identity map. // TODO avoid using this
    bool heavyLines[NOP];
    int highestWeight = 0;
    for (int i = 0; i < NOP; ++i) {
        if (set.weight[i] > highestWeight) {
            highestWeight = set.weight[i];
            for (bool &heavyLine : heavyLines) {
                heavyLine = false;
            }
            heavyLines[i] = true;
        } else if (set.weight[i] == highestWeight) {
            heavyLines[i] = true;
        }
    }
    for (int l = 0; l < NOP; ++l) {
        if (heavyLines[l]) {
            for (int i = 0; i < SGS_SIZE; ++i) {
                int index = hlmList[l][i];
                const Matrix &invMap = list[index].adjoint();
                bool kado1 = set.present[invMap(Point::list[0]).index()];
                bool kado2 = set.present[invMap(Point::list[1]).index()];
                if (kado1 && kado2) { // Points 0 and 1 are always in canonicalSet //TODO: Point 2 toch ook?
                    for (int j = 2; j < NOP; ++j) {
                        bool kado = set.present[invMap(Point::list[j]).index()];
                        if (kado && !canonicalSet.present[j]) {
                            //invMap induces a higher lexicographical ordering. break the loop that iterates over all points.
                            bestMap = index;
                            if (j > NOP - set.size -  1) { // if this holds, then the implementation below has to map less points than apply does.
                                // TODO avoid these ifs
                                if (!canonicalSet.present[j]) {
                                    canonicalSet.add(j);
                                }
                                for (int k = j + 1; k < NOP; ++k) {
                                    if (set.present[invMap(Point::list[k]).index()]) {
                                        if (!canonicalSet.present[k]) {
                                            canonicalSet.add(k); // might add a point that was already internal
                                        }
                                    } else {
                                        if (canonicalSet.present[k]) {
                                            canonicalSet.remove(k); // might delete a point that was already external
                                        }
                                    }
                                }
                                canonicalSet.updateInvariants();
                            } else {
                                canonicalSet = set.apply(list[index]);
                            }
                            break;
                        } else if (!kado && canonicalSet.present[j]) {
                            //map induces a lower lexicographical ordering
                            break;
                        }
                    }
                }
            }
        }
    }
    return bestMap;
}


#endif //BSETS2_GROUP_H
