// Note: this file is part of a header-only library. See Readme.md

#ifndef BSETS2_POINTSET_H
#define BSETS2_POINTSET_H

#include "Point.hpp"
#include "Matrix.hpp"

////////////////////////////////////////
// DECLARATIONS
////////////////////////////////////////

class PointSet {

public:
    // is a point present in this set?
    std::array<bool, NOP> present{false,};

    // size of the set
    int size = 0;

    // invariant value for each point of the set
    std::array<int, NOP> invariants{0,};

    // weights of lines blocked by this set
    std::array<int, NOP> weight{0,};

    PointSet() = default;

public:
    // Add a point. Should only be called when the point is not already in the set. This is not checked.
    // Adjust weights but not invariants
    void add(int pt);

    // Add a point and checks whether the resulting set can still be completed to a minimal blocking set
    bool addAndCheck(int pt);

    // Remove a point that was previously added. Should only be called when the point is already in the set. This is not checked.
    // Adjust weights but not invariants
    void remove(int pt);

    // update the invariants array to reflect the current set
    void updateInvariants();

    // ship a representation of this set to the given stream
    void ship(std::ostream &ostream) const;

    // set equality (uses present[])
    bool operator==(const PointSet &other) const;

    bool operator!=(const PointSet &other) const;

    // is subset of other? (Currently only used for debugging)
    bool operator<=(const PointSet &other) const;

    // a blocking set is a set of points that touches every line, but does not contain an entire line.
    bool isBlockingSet() const;

    // computes an invariant for the entire set. Only call after call to updateInvariants
    int computeInvariant() const;

    // Returns a point with a unique invariant value, or -1 if no such point exists
    int hasUniqueInvariant() const;

    // Computes the image of this point by the given matrix. The result has updated invariants.
    PointSet apply(const Matrix &matrix) const;

    // Is this set stabilized bij the given matrix? Only call after call to updateInvariants
    bool isFixedBy(const Matrix &matrix) const;

    // Check whether the given point has the largest invariant value. If so, return the number of points
    // that have the same in variant value, if not return 0.
    int checkLargestInvariant(int pt) const;

    // Check whether the given point is the point with the largest unique invariant
    // Returns 1 if that is the case, returns -1 if another point has the largest unique invariant and returns 0
    // if no invariant value is unique
    int checkLargestUniqueInvariant(int pt) const;

private:

    // computes the invariant for the given point
    int computeInvariant(int pt) const;

    // a point is redundant if all lines through it have weight > 2
    bool isRedundant(int pt) const;

    //static
    static const int invarHash[Q + 2]; // invarHash[0] should never be used.

public:
    // Debugging only
    void printPointSet() const;

    void printInvariants() const;

    void printWeights() const;

};

////////////////////////////////////////
// DEFINITIONS
////////////////////////////////////////

// Add a point. Should only be called when the point is not already in the set. This is not checked.
// Adjusts weights but not invariants
void PointSet::add(int pt) {
    present[pt] = true;
    size++;
    for (const auto &ln : Point::pIl[pt]) {
        weight[ln]++;
    }
}

// Remove a point. Should only be called when the point is already in the set. This is not checked.
// Adjusts weights but not invariants
void PointSet::remove(int pt) {
    present[pt] = false;
    size--;
    for (const auto &ln : Point::pIl[pt]) {
        weight[ln]--;
    }
}

// computes the invariant for the given point
int PointSet::computeInvariant(int pt) const {
    int value = 0;
    for (const int &ln : Point::pIl[pt]) {
        value += invarHash[weight[ln]];
    }
    return value;
}

void PointSet::updateInvariants() {
    for (int i = 0; i < NOP; ++i) {
        if (present[i]) {
            invariants[i] = computeInvariant(i);
        } else {
            invariants[i] = 0;
        }
    }
}

int PointSet::computeInvariant() const {
    int result = 0;
    for (int i = 0; i < NOP; ++i) {
        if (present[i]) {
            result += invariants[i] * invariants[i]; // integer overflow may occur, but that is not a problem
        }
    }
    return result;
}

bool PointSet::operator==(const PointSet &other) const {
    // checking the present arrays
    for (int i = 0; i < NOP; ++i) {
        if (present[i] != other.present[i]) {
            return false;
        }
    }
    return true;
}

bool PointSet::operator!=(const PointSet &other) const {
    for (int i = 0; i < NOP; ++i) {
        if (present[i] != other.present[i]) {
            return true;
        }
    }
    return false;
}

bool PointSet::addAndCheck(int pt) {
    add(pt);

    if (size <= Q) {
        return true;
    }

    // should not create a full line
    for (const auto &ln : Point::pIl[pt]) {
        if (weight[ln] > Q) {
            return false;
        }
    }

    // should not create a redundant point
    if (size <= Q + 1) {
        return true;
    }

    for (int i = 0; i < NOP; ++i) {
        // TODO could be improved slightly by running only over those points on a line through pt that has weight at least 2
        // probably not worth it
        if (present[i] && isRedundant(i)) {
            return false;
        }
    }

    return true;
}

// a point is redundant if all lines through it have weight at least 2
bool PointSet::isRedundant(int pt) const {
    for (const auto &ln : Point::pIl[pt]) {
        if (weight[ln] < 2) {
            return false;
        }
    }
    return true;
}

bool PointSet::isFixedBy(const Matrix &matrix) const {
    for (int pt = 0; pt < NOP; pt++) {
        if (present[pt]) {
            int image = matrix(Point::list[pt]).index();
            if (invariants[image] != invariants[pt] || present[pt] != present[image]) {
                return false;
            }
        }
    }
    return true;
}

// Returns a point with a unique invariant value, or -1 if no such point exists
int PointSet::hasUniqueInvariant() const {
    std::array<bool, NOP> todo = present; // makes a copy!
    for (int i = 0; i < NOP; ++i) {
        if (todo[i]) {
            bool found = false;
            for (int j = i + 1; j < NOP; ++j) {
                if (todo[j] && invariants[i] == invariants[j]) {
                    found = true;
                    todo[j] = false;
                }
            }
            if (!found) {
                return i;
            }
        }
    }
    return -1;
}

// Check whether the given point has the largest invariant value. If so, return the number of points
// that have the same in variant value, if not return 0.
int PointSet::checkLargestInvariant(int pt) const {
    int count = 0;
    int value = invariants[pt];
    for (int i = 0; i < NOP; ++i) {
        if (present[i]) {
            int otherValue = invariants[i];
            if (otherValue > value) {
                return 0; // pt does NOT have the largest invariant
            } else if (otherValue == value) {
                count++;
            }
        }
    }
    return count;
}

// Check whether the given point is the point with the largest unique invariant
// Returns 1 if that is the case, returns -1 if another point has the largest unique invariant or the given point
// has not a unique invariant value and returns 0 if no invariant value is unique
int PointSet::checkLargestUniqueInvariant(int pt) const {
    std::array<bool, NOP> todo = present; // makes a copy
    // this array keeps track of which points should still be visited,
    // after the loop, only for the unique invariants the value remains true
    int nrOfUniqueInvars = 0;
    for (int i = 0; i < NOP; ++i) {
        if (todo[i]) {
            bool found = false;
            for (int j = i + 1; j < NOP; ++j) {
                if (todo[j] && invariants[i] == invariants[j]) {
                    found = true;
                    todo[j] = false;
                }
            }
            if (found) {
                todo[i] = false;
            } else {
                // i has a unique invariant value
                nrOfUniqueInvars++;
                if (invariants[i] > invariants[pt]) {
                    return -1;
                }
            }
        }
    }
    if (nrOfUniqueInvars == 0) {
        return 0;
    } else if (!todo[pt]) {
        return -1;
    } else {
        return 1;
    }
}

PointSet PointSet::apply(const Matrix &matrix) const {
    PointSet image;
    for (int i = 0; i < NOP; ++i) {
        if (present[i]) {
            image.add(matrix(Point::list[i]).index());
        }
    }
    image.updateInvariants();
    return image;
}


bool PointSet::isBlockingSet() const {
    for (int i = 0; i < NOP; ++i) {
        if (weight[i] == 0 || weight[i] == Q + 1) {
            return false;
        }
    }
    return true;
}


void PointSet::ship(std::ostream &ostream) const {
    for (int i = 0; i < NOP; ++i) {
        if (present[i]) {
            ostream << i << "-";
        }
    }
    ostream << std::endl;
}

/////////////////////////////////////////////
// DEBUGGING TOOLS
/////////////////////////////////////////////

void PointSet::printPointSet() const {
    std::cout << "PointSet: [";
    for (int i = 0; i < NOP; ++i) {
        if (present[i]) {
            std::cout << " 1";
        } else { std::cout << " 0"; }
    }
    std::cout << "]" << std::endl;
}

void PointSet::printInvariants() const {
    std::cout << "Invariants: [";
    for (int i = 0; i < NOP; ++i) {
        std::cout << " " << invariants[i];
    }
    std::cout << "]" << std::endl;
}

void PointSet::printWeights() const {
    std::cout << "Weights: [";
    for (int i = 0; i < NOP; ++i) {
        std::cout << " " << weight[i];
    }
    std::cout << "]" << std::endl;
}

bool PointSet::operator<=(const PointSet &other) const {
    for (int i=0; i < NOP; i++) {
        if (present[i] && !other.present[i]) {
            return false;
        }
    }
    return true;
}




// initializations
#if ORDER == 3

const int PointSet::invarHash[Q+2] { 0,1,10,100,1000 };

#elif ORDER == 4

const int PointSet::invarHash[Q+2] { 0,1,10,100,1000,10000 };

#elif ORDER == 5

const int PointSet::invarHash[Q + 2]{0, 1, 10, 100, 1000, 10000, 100000};

#elif ORDER == 7

const int PointSet::invarHash[Q + 2]{0, 1, 10, 100, 1000, 10000, 100000, 1000000, 10000000};

#elif ORDER == 8

const int PointSet::invarHash[Q+2] { 0,1,10,100,1000,10000,100000, 1000000, 10000000, 100000000 };

#elif ORDER==9

const int PointSet::invarHash[Q+2] { 0,1,10,100,1000,10000,100000, 1000000, 10000000, 100000000, 1000000000 };

#endif


#endif //BSETS2_POINTSET_H
